"""An ELK reporter network."""

from .losses import ccs_squared_loss, js_loss
from copy import deepcopy
from dataclasses import dataclass
from pathlib import Path
from sklearn.metrics import roc_auc_score
from simple_parsing.helpers import Serializable
from torch.nn.functional import binary_cross_entropy as bce
from typing import cast, Literal, NamedTuple, Optional, Union
import math
import torch
import torch.nn as nn


class EvalResult(NamedTuple):
    """The result of evaluating a reporter on a dataset.

    The `.score()` function of a reporter returns an instance of this class,
    which contains the loss, accuracy, calibrated accuracy, and AUROC.
    """

    loss: float
    acc: float
    cal_acc: float
    auroc: float


@dataclass
class ReporterConfig(Serializable):
    """
    Args:
        activation: The activation function to use. Defaults to GELU.
        bias: Whether to use a bias term in the linear layers. Defaults to True.
        device: The device to use. Defaults to None, which means "current device".
        hidden_size: The number of hidden units in the MLP. Defaults to None.
            By default, use an MLP expansion ratio of 4/3. This ratio is used by
            Tucker et al. (2022) <https://arxiv.org/abs/2204.09722> in their 3-layer
            MLP probes. We could also use a ratio of 4, imitating transformer FFNs,
            but this seems to lead to excessively large MLPs when num_layers > 2.
        init: The initialization scheme to use. Defaults to "zero".
        loss: The loss function to use. Defaults to "squared".
        num_layers: The number of layers in the MLP. Defaults to 1.
        pre_ln: Whether to include a LayerNorm module before the first linear
            layer. Defaults to False.
        supervised_weight: The weight of the supervised loss. Defaults to 0.0.
    """

    activation: Literal["gelu", "relu", "swish"] = "gelu"
    bias: bool = True
    hidden_size: Optional[int] = None
    init: Literal["default", "spherical", "zero"] = "default"
    loss: Literal["js", "squared"] = "squared"
    num_layers: int = 1
    pre_ln: bool = False
    seed: int = 42
    supervised_weight: float = 0.0


@dataclass
class OptimConfig(Serializable):
    """
    Args:
        lr: The learning rate to use. Ignored when `optimizer` is `"lbfgs"`.
            Defaults to 1e-2.
        num_epochs: The number of epochs to train for. Defaults to 1000.
        num_tries: The number of times to try training the reporter. Defaults to 10.
        optimizer: The optimizer to use. Defaults to "adam".
        weight_decay: The weight decay or L2 penalty to use. Defaults to 0.01.
    """

    lr: float = 1e-2
    num_epochs: int = 1000
    num_tries: int = 10
    optimizer: Literal["adam", "lbfgs"] = "lbfgs"
    weight_decay: float = 0.01


class Reporter(nn.Module):
    """An ELK reporter network.

    Args:
        in_features: The number of input features.
        cfg: The reporter configuration.
    """

    def __init__(
        self, in_features: int, cfg: ReporterConfig, device: Optional[str] = None
    ):
        super().__init__()

        hidden_size = cfg.hidden_size or 4 * in_features // 3

        self.probe = nn.Sequential(
            nn.Linear(
                in_features,
                1 if cfg.num_layers < 2 else hidden_size,
                bias=cfg.bias,
                device=device,
            ),
        )
        if cfg.pre_ln:
            self.probe.insert(0, nn.LayerNorm(in_features, elementwise_affine=False))

        act_cls = {
            "gelu": nn.GELU,
            "relu": nn.ReLU,
            "swish": nn.SiLU,
        }[cfg.activation]

        for i in range(1, cfg.num_layers):
            self.probe.append(act_cls())
            self.probe.append(
                nn.Linear(
                    hidden_size,
                    1 if i == cfg.num_layers - 1 else hidden_size,
                    bias=cfg.bias,
                    device=device,
                )
            )

        self.init = cfg.init
        self.device = device
        self.unsupervised_loss = js_loss if cfg.loss == "js" else ccs_squared_loss
        self.supervised_weight = cfg.supervised_weight

    def reset_parameters(self):
        """Reset the parameters of the probe.

        If init is "spherical", use the spherical initialization scheme.
        If init is "default", use the default PyTorch initialization scheme for
        nn.Linear (Kaiming uniform).
        If init is "zero", initialize all parameters to zero.
        """
        if self.init == "spherical":
            # Mathematically equivalent to the unusual initialization scheme used in
            # the original paper. They sample a Gaussian vector of dim in_features + 1,
            # normalize to the unit sphere, then add an extra all-ones dimension to the
            # input and compute the inner product. Here, we use nn.Linear with an
            # explicit bias term, but use the same initialization.
            assert len(self.probe) == 1, "Only linear probes can use spherical init"
            probe = cast(nn.Linear, self.probe[0])  # Pylance gets the type wrong here

            theta = torch.randn(1, probe.in_features + 1, device=probe.weight.device)
            theta /= theta.norm()
            probe.weight.data = theta[:, :-1]
            probe.bias.data = theta[:, -1]

        elif self.init == "default":
            for layer in self.probe:
                if isinstance(layer, nn.Linear):
                    layer.reset_parameters()
        elif self.init == "zero":
            for param in self.parameters():
                param.data.zero_()
        else:
            raise ValueError(f"Unknown init: {self.init}")

    # TODO: These methods will do something fancier in the future
    @classmethod
    def load(cls, path: Union[Path, str]):
        """Load a reporter from a file."""
        return torch.load(path)

    def save(self, path: Union[Path, str]):
        # TODO: Save separate JSON and PT files for the reporter.
        torch.save(self, path)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Return the raw score output of the probe on `x`."""
        return self.probe(x).squeeze(-1)

    def loss(
        self,
        logit0: torch.Tensor,
        logit1: torch.Tensor,
        labels: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        """Return the loss of the reporter on the contrast pair (x0, x1).

        Args:
            logit0: The raw score output of the reporter on x0.
            logit1: The raw score output of the reporter on x1.
            labels: The labels of the contrast pair. Defaults to None.

        Returns:
            loss: The loss of the reporter on the contrast pair (x0, x1).

        Raises:
            ValueError: If `supervised_weight > 0` but `labels` is None.
        """
        loss = self.unsupervised_loss(logit0, logit1)

        # If labels are provided, use them to compute a supervised loss
        if labels is not None:
            num_labels = len(labels)
            assert num_labels <= len(logit0), "Too many labels provided"
            p0 = logit0[:num_labels].sigmoid()
            p1 = logit1[:num_labels].sigmoid()

            alpha = self.supervised_weight
            preds = p0.add(1 - p1).mul(0.5).squeeze(-1)
            bce_loss = bce(preds, labels.type_as(preds))
            loss = alpha * bce_loss + (1 - alpha) * loss

        elif self.supervised_weight > 0:
            raise ValueError(
                "Supervised weight > 0 but no labels provided to compute loss"
            )

        return loss

    def validate_data(self, data):
        """Validate that the data's shape is valid."""
        assert len(data) == 2 and data[0].shape == data[1].shape

    def fit(
        self,
        contrast_pair: tuple[torch.Tensor, torch.Tensor],
        labels: Optional[torch.Tensor] = None,
        cfg: OptimConfig = OptimConfig(),
    ) -> float:
        """Fit the probe to the contrast pair (x0, x1).

        Args:
            contrast_pair: A tuple of tensors, (x0, x1), where x0 and x1 are the
                contrastive representations.
            labels: The labels of the contrast pair. Defaults to None.
            lr: The learning rate for Adam. Defaults to 1e-2.
            num_epochs: The number of epochs to train for. Defaults to 1000.
            num_tries: The number of times to repeat the procedure. Defaults to 10.
            optimizer: The optimizer to use. Defaults to "adam".
            verbose: Whether to print out information at each step. Defaults to False.
            weight_decay: The weight decay for Adam. Defaults to 0.01.

        Returns:
            best_loss: The best loss obtained.

        Raises:
            ValueError: If `optimizer` is not "adam" or "lbfgs".
            RuntimeError: If the best loss is not finite.
        """
        self.validate_data(contrast_pair)

        # Record the best acc, loss, and params found so far
        best_loss = torch.inf
        best_state: dict[str, torch.Tensor] = {}  # State dict of the best run
        x0, x1 = contrast_pair

        for _ in range(cfg.num_tries):
            self.reset_parameters()

            if cfg.optimizer == "lbfgs":
                loss = self.train_loop_lbfgs(x0, x1, labels, cfg)
            elif cfg.optimizer == "adam":
                loss = self.train_loop_adam(x0, x1, labels, cfg)
            else:
                raise ValueError(f"Optimizer {cfg.optimizer} is not supported")

            if loss < best_loss:
                best_loss = loss
                best_state = deepcopy(self.state_dict())

        if not math.isfinite(best_loss):
            raise RuntimeError("Got NaN/infinite loss during training")

        self.load_state_dict(best_state)
        return best_loss

    @torch.no_grad()
    def score(
        self,
        contrast_pair: tuple[torch.Tensor, torch.Tensor],
        labels: torch.Tensor,
    ) -> EvalResult:
        """Score the probe on the contrast pair (x0, x1).

        Args:
            contrast_pair: A tuple of tensors, (x0, x1), where x0 and x1 are the
                contrastive representations.
            labels: The labels of the contrast pair.

        Returns:
            an instance of EvalResult containing the loss, accuracy, calibrated
                accuracy, and AUROC of the probe on the contrast pair (x0, x1).
        """

        self.validate_data(contrast_pair)

        logit0, logit1 = map(self, contrast_pair)
        p0, p1 = logit0.sigmoid(), logit1.sigmoid()
        pred_probs = 0.5 * (p0 + (1 - p1))

        # Calibrated accuracy
        cal_thresh = pred_probs.float().quantile(labels.float().mean())
        cal_preds = pred_probs.gt(cal_thresh).squeeze(1).to(torch.int)
        raw_preds = pred_probs.gt(0.5).squeeze(1).to(torch.int)

        auroc = float(roc_auc_score(labels.cpu(), pred_probs.cpu()))
        cal_acc = cal_preds.eq(labels.reshape(-1)).float().mean()
        raw_acc = raw_preds.eq(labels.reshape(-1)).float().mean()

        return EvalResult(
            loss=self.loss(logit0, logit1).item(),
            acc=torch.max(raw_acc, 1 - raw_acc).item(),
            cal_acc=torch.max(cal_acc, 1 - cal_acc).item(),
            auroc=max(auroc, 1 - auroc),
        )

    def train_loop_adam(
        self,
        x0,
        x1,
        labels: Optional[torch.Tensor],
        cfg: OptimConfig,
    ) -> float:
        """Adam train loop, returning the final loss. Modifies params in-place."""

        optimizer = torch.optim.AdamW(
            self.parameters(), lr=cfg.lr, weight_decay=cfg.weight_decay
        )

        loss = torch.inf
        for _ in range(cfg.num_epochs):
            optimizer.zero_grad()

            loss = self.loss(self(x0), self(x1), labels)
            loss.backward()
            optimizer.step()

        return float(loss)

    def train_loop_lbfgs(
        self, x0, x1, labels: Optional[torch.Tensor], cfg: OptimConfig
    ) -> float:
        """LBFGS train loop, returning the final loss. Modifies params in-place."""

        optimizer = torch.optim.LBFGS(
            self.parameters(),
            line_search_fn="strong_wolfe",
            max_iter=cfg.num_epochs,
            tolerance_change=torch.finfo(x0.dtype).eps,
            tolerance_grad=torch.finfo(x0.dtype).eps,
        )
        # Raw unsupervised loss, WITHOUT regularization
        loss = torch.inf

        def closure():
            nonlocal loss
            optimizer.zero_grad()

            loss = self.loss(self(x0), self(x1), labels)
            regularizer = 0.0

            # We explicitly add L2 regularization to the loss, since LBFGS
            # doesn't have a weight_decay parameter
            for param in self.parameters():
                regularizer += cfg.weight_decay * param.norm() ** 2 / 2

            regularized = loss + regularizer
            regularized.backward()

            return float(regularized)

        optimizer.step(closure)
        return float(loss)



from .typing import assert_type
from datasets import ClassLabel, Dataset, DatasetDict, Features, concatenate_datasets
from random import Random
from typing import Optional
import numpy as np
import torch


def compute_class_balance(
    dataset: Dataset, label_column: Optional[str] = None
) -> np.ndarray:
    """Compute the class balance of a `Dataset`."""

    features = dataset.features
    name = dataset.info.builder_name
    if label_column is None:
        label_column = infer_label_column(dataset.features)
    elif label_column not in features:
        raise ValueError(f"{name} has no column '{label_column}'")

    num_classes = getattr(features[label_column], "num_classes", 0)
    class_sizes = np.bincount(dataset[label_column], minlength=num_classes)

    if not np.all(class_sizes > 0):
        missing = np.flatnonzero(class_sizes == 0).tolist()
        raise ValueError(f"{name} has missing classes: {missing}")

    return class_sizes


def get_columns_all_equal(dataset: DatasetDict) -> list[str]:
    """Get columns of a `DatasetDict`, asserting all splits have the same columns."""
    pivot, *rest = dataset.column_names.values()

    if not all(cols == pivot for cols in rest):
        raise ValueError("All splits must have the same columns")

    return pivot


def held_out_split(dataset: DatasetDict) -> Dataset:
    """Return the validation set if it exits, otherwise the test set."""
    if "validation" in dataset:
        return dataset["validation"]
    elif "test" in dataset:
        return dataset["test"]
    else:
        raise ValueError("No validation or test split found")


def infer_label_column(features: Features) -> str:
    """Return the unique `ClassLabel` column in a `Dataset`.

    Returns:
        The name of the unique label column.
    Raises:
        ValueError: If there are no `ClassLabel` columns, or if there are multiple.
    """
    label_cols = [
        col for col, dtype in features.items() if isinstance(dtype, ClassLabel)
    ]
    if not label_cols:
        raise ValueError("Dataset has no label column")
    elif len(label_cols) > 1:
        raise ValueError(
            f"Dataset has multiple label columns {label_cols}; specify "
            f"label_column to disambiguate"
        )
    else:
        return assert_type(str, label_cols[0])


def undersample(
    dataset: Dataset, rng: Random, label_column: Optional[str] = None
) -> Dataset:
    """Undersample a `Dataset` to the smallest class size."""
    label_column = label_column or infer_label_column(dataset.features)
    class_sizes = compute_class_balance(dataset, label_column)
    smallest_size = class_sizes.min()

    # First group the active split by class
    strata = (
        dataset.filter(lambda ex: ex[label_column] == i)
        for i in range(len(class_sizes))
    )
    # Then randomly sample `smallest_size` examples from each class and merge
    strata = [
        stratum.select(rng.sample(range(len(stratum)), k=smallest_size))
        for stratum in strata
    ]
    dataset = assert_type(Dataset, concatenate_datasets(strata))

    # Sanity check that we successfully balanced the classes
    class_sizes = np.bincount(dataset[label_column], minlength=len(class_sizes))
    assert np.all(class_sizes == smallest_size)

    return dataset


def float32_to_int16(x: torch.Tensor) -> torch.Tensor:
    """Converts float32 to float16, then reinterprets as int16."""
    return x.type(torch.float16).view(torch.int16)


def int16_to_float32(x: torch.Tensor) -> torch.Tensor:
    """Converts int16 to float16, then reinterprets as float32."""
    return x.view(torch.float16).type(torch.float32)



"""Port of `jax.tree_util` to PyTorch.

In JAX, "pytrees" are nested collections (lists, tuples, dicts) of tensors which can be
mapped over. See <https://jax.readthedocs.io/en/latest/pytrees.html>.
"""

from typing import Callable, Mapping, TypeVar


TreeType = TypeVar("TreeType")


def pytree_map(func: Callable, tree: TreeType) -> TreeType:
    """
    Recursively apply a function to all objects in a pytree, returning the results
    in a new pytree with the same structure.

    Examples:
    >>> pytree_map(lambda x: x + 1, {"x": 7, "y": 42})
    {'x': 8, 'y': 43}
    """
    # Recursive case
    if isinstance(tree, Mapping):
        return {k: pytree_map(func, v) for k, v in tree.items()}  # type: ignore

    if isinstance(tree, list):
        return [pytree_map(func, v) for v in tree]  # type: ignore

    if isinstance(tree, tuple):
        return tuple(pytree_map(func, v) for v in tree)  # type: ignore

    # Stopping condition
    return func(tree)



import math
import random


def stochastic_round_constrained(x: list[float], rng: random.Random) -> list[int]:
    """Stochastic rounding under integer constraints.

    Given a list of floats which sum to an integer, stochastically round each float to
    an integer while maintaining the same sum. The expectation of the rounded values
    should be (nearly?) equal to the original values.

    Inspired by the algorithm described in https://arxiv.org/abs/1501.00014. Instead of
    deterministically allocating the total shortfall to the elements with the largest
    fractional parts, we randomly allocate it to elements with probability proportional
    to their fractional parts.
    """

    total = sum(x)
    assert math.isclose(total, round(total)), "Total must be an integer"

    rounded = [math.floor(x_) for x_ in x]
    fractional_parts = [x_ - r_ for x_, r_ in zip(x, rounded)]

    # Randomly choose where to allocate the shortfall. Note that currently we are
    # allocating the entire shortfall to a single element, but we could allocate
    # fractions of the shortfall to multiple elements. This would be lower variance
    # but requires an RNG that supports weighted sampling without replacement- this
    # exists in NumPy but not in the Python standard library so we'd need to switch
    # this function to use NumPy (and maintain a NumPy RNG).
    if any(fractional_parts):
        (index,) = rng.choices(range(len(x)), weights=fractional_parts)
        rounded[index] += round(sum(fractional_parts))

    return rounded



"""Main training loop."""

from ..extraction import extract, ExtractionConfig
from ..files import elk_reporter_dir, memorably_named_dir
from ..utils import assert_type, held_out_split, select_usable_devices, int16_to_float32
from .preprocessing import normalize
from .reporter import OptimConfig, Reporter, ReporterConfig
from dataclasses import dataclass
from datasets import DatasetDict
from functools import partial
from pathlib import Path
from simple_parsing import Serializable
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import accuracy_score, roc_auc_score
from torch import Tensor
from tqdm.auto import tqdm
from typing import cast, Literal, Optional
import csv
import numpy as np
import os
import pickle
import random
import torch
import torch.multiprocessing as mp


@dataclass
class RunConfig(Serializable):
    """Full specification of a reporter training run.

    Args:
        data: Config specifying hidden states on which the reporter will be trained.
        net: Config for building the reporter network.
        optim: Config for the `.fit()` loop.
    """

    data: ExtractionConfig
    net: ReporterConfig
    optim: OptimConfig

    label_frac: float = 0.0
    max_gpus: int = -1
    normalization: Literal["legacy", "none", "elementwise", "meanonly"] = "meanonly"
    skip_baseline: bool = False


def train_reporter(
    cfg: RunConfig,
    dataset: DatasetDict,
    out_dir: Path,
    layer: int,
    devices: list[str],
    world_size: int = 1,
):
    """Train a single reporter on a single layer."""

    # Reproducibility
    seed = cfg.net.seed + layer
    np.random.seed(seed)
    random.seed(seed)
    torch.manual_seed(seed)

    rank = os.getpid() % world_size
    device = devices[rank]

    # Note: currently we're just upcasting to float32 so we don't have to deal with
    # grad scaling (which isn't supported for LBFGS), while the hidden states are
    # saved in float16 to save disk space. In the future we could try to use mixed
    # precision training in at least some cases.
    with dataset.formatted_as("torch", device=device, dtype=torch.int16):
        train, val = dataset["train"], held_out_split(dataset)
        train_labels = cast(Tensor, train["label"])
        val_labels = cast(Tensor, val["label"])

        train_h, val_h = normalize(
            int16_to_float32(assert_type(Tensor, train[f"hidden_{layer}"])),
            int16_to_float32(assert_type(Tensor, val[f"hidden_{layer}"])),
            method=cfg.normalization,
        )
        x0, x1 = train_h.unbind(dim=-2)
        val_x0, val_x1 = val_h.unbind(dim=-2)

    reporter = Reporter(x0.shape[-1], cfg.net, device=device)
    if cfg.label_frac:
        num_labels = round(cfg.label_frac * len(train_labels))
        labels = train_labels[:num_labels]
    else:
        labels = None

    train_loss = reporter.fit((x0, x1), labels, cfg.optim)
    val_result = reporter.score(
        (val_x0, val_x1),
        val_labels,
    )

    lr_dir = out_dir / "lr_models"
    reporter_dir = out_dir / "reporters"

    lr_dir.mkdir(parents=True, exist_ok=True)
    reporter_dir.mkdir(parents=True, exist_ok=True)
    stats = [layer, train_loss, *val_result]

    if not cfg.skip_baseline:
        train_labels_aug = torch.cat([train_labels, 1 - train_labels]).cpu()
        val_labels_aug = torch.cat([val_labels, 1 - val_labels]).cpu()

        # TODO: Once we implement cross-validation for CCS, we should benchmark
        # against LogisticRegressionCV here.
        X = torch.cat([x0, x1]).cpu().squeeze()
        lr_model = LogisticRegression(max_iter=10_000)
        lr_model.fit(X.view(-1, X.shape[-1]), train_labels_aug)

        X_val = torch.cat([val_x0, val_x1]).cpu().squeeze()
        lr_preds = lr_model.predict_proba(X_val)[:, 1]
        lr_acc = accuracy_score(val_labels_aug, lr_preds > 0.5)
        lr_auroc = roc_auc_score(val_labels_aug, lr_preds)

        stats += [lr_auroc, lr_acc]
        with open(lr_dir / f"layer_{layer}.pkl", "wb") as file:
            pickle.dump(lr_model, file)

    with open(reporter_dir / f"layer_{layer}.pt", "wb") as file:
        torch.save(reporter, file)

    return stats


def train(cfg: RunConfig, out_dir: Optional[Path] = None):
    # Extract the hidden states first if necessary
    ds = extract(cfg.data, max_gpus=cfg.max_gpus)

    if out_dir is None:
        out_dir = memorably_named_dir(elk_reporter_dir())
    else:
        out_dir.mkdir(parents=True, exist_ok=True)

    # Print the output directory in bold with escape codes
    print(f"Saving results to \033[1m{out_dir}\033[0m")

    with open(out_dir / "cfg.yaml", "w") as f:
        cfg.dump_yaml(f)

    devices = select_usable_devices(cfg.max_gpus)
    num_devices = len(devices)

    cols = ["layer", "train_loss", "loss", "acc", "cal_acc", "auroc"]
    if not cfg.skip_baseline:
        cols += ["lr_auroc", "lr_acc"]

    layers = [
        int(feat[len("hidden_") :])
        for feat in ds["train"].features
        if feat.startswith("hidden_")
    ]
    # Train reporters for each layer in parallel
    with mp.Pool(num_devices) as pool, open(out_dir / "eval.csv", "w") as f:
        fn = partial(
            train_reporter, cfg, ds, out_dir, devices=devices, world_size=num_devices
        )
        writer = csv.writer(f)
        writer.writerow(cols)

        mapper = pool.imap_unordered if num_devices > 1 else map
        for i, *stats in tqdm(mapper(fn, layers), total=len(layers)):
            writer.writerow([i] + [f"{s:.4f}" for s in stats])



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as readme_file:
    readme = readme_file.read()

requirements = [
    "black<=21.12b0",
    "datasets>=1.7.0",
    "flake8",
    "isort==5.8.0",
    "pytest",
    "pyyaml>=5",
    "streamlit==0.82",
    "jinja2",
    "plotly",
    "requests",
    "pandas",
    ##############################################################
    # Dependencies in this section are added for specific datasets
    ##############################################################
    "py7zr",
    ##############################################################
    # End of dataset-specific dependencies
    ##############################################################
]

setup(
    name='promptsource',
    version='0.2.3',
    url='https://github.com/bigscience-workshop/promptsource.git',
    author='BigScience - Prompt Engineering Working Group',
    author_email='sbach@cs.brown.edu,victor@huggingface.co',
    python_requires='>=3.7',
    install_requires=requirements,
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    description='An Integrated Development Environment and Repository for Natural Language Prompts.',
    packages=find_packages(),
    license="Apache Software License 2.0",
    long_description=readme,
    long_description_content_type="text/markdown",
    package_data={"": [
        "templates/*/*.yaml",
        "templates/*/*/*.yaml",
    ]}
)



"""Helper functions for dealing with files."""

from pathlib import Path
import json
import os
import random


def elk_reporter_dir() -> Path:
    """Return the directory where reporter checkpoints and logs are stored."""
    env_dir = os.environ.get("ELK_DIR", None)
    if env_dir is None:
        log_dir = Path.home() / "elk-reporters"
    else:
        log_dir = Path(env_dir)

    log_dir.mkdir(parents=True, exist_ok=True)
    return log_dir


def memorably_named_dir(parent: Path):
    """Return a memorably-named cached directory of the form 'goofy-goodall'."""
    resource_dir = Path(__file__).parent / "resources"

    with open(resource_dir / "adjectives.json", "r") as f:
        adjectives = json.load(f)
    with open(resource_dir / "names.json", "r") as f:
        names = json.load(f)

    sub_dir = "."
    while parent.joinpath(sub_dir).exists():
        adj = random.choice(adjectives)
        name = random.choice(names)
        sub_dir = f"{adj}-{name}"

    out_dir = parent / sub_dir
    out_dir.mkdir(parents=True, exist_ok=True)
    return out_dir



from elk.math import stochastic_round_constrained
from hypothesis import given, strategies as st
from random import Random
import math
import numpy as np
import pytest


@given(
    # Let Hypothesis generate the number of floats...
    st.integers(min_value=1, max_value=100),
    # ...and the total sum of the floats
    st.integers(min_value=1, max_value=int(np.finfo(np.float32).max)),
)
@pytest.mark.cpu
def test_stochastic_rounding(num_parts: int, total: int):
    # Randomly sample the breakdown of the total into floats
    rng = np.random.default_rng(42)
    x = rng.dirichlet(np.ones(num_parts), size=1) * total

    # Stochastically round the floats
    rounded = stochastic_round_constrained(x[0].tolist(), Random(42))

    # Check that the rounded floats sum to the total
    assert math.isclose(sum(rounded), total)

    # Check that the rounded floats are never smaller than np.floor(x)
    assert all(math.floor(x_) <= r_ for x_, r_ in zip(x[0], rounded))

    # TODO: Check that the rounded floats are never larger than np.ceil(x),
    # once our implementation actually has this property.



"""Loss functions for training reporters."""

from torch import Tensor
import math
import torch


def H(p: Tensor) -> Tensor:
    """Entropy of Bernoulli distribution(s) with success probability `p`."""
    return torch.nn.functional.binary_cross_entropy(p, p)


def js_loss(
    logit0: Tensor,
    logit1: Tensor,
    confidence: float = 0.0,
    base: float = 2.0,
) -> Tensor:
    """Consistency and confidence loss based on the Jensen-Shannon divergence.

    Note that by default we use the base 2 logarithm, so the value is measured in bits.
    This ensures the divergence is in the range [0, 1]."""
    p0, neg_p1 = logit0.sigmoid(), 1 - logit1.sigmoid()
    nats = (1 + confidence) * H((p0 + neg_p1) / 2) - (H(p0) + H(neg_p1)) / 2
    return nats / math.log(base)


def ccs_squared_loss(logit0: Tensor, logit1: Tensor) -> Tensor:
    """CCS loss from original paper, with squared differences between probabilities.

    The loss is symmetric, so it doesn't matter which argument is the original and
    which is the negated proposition.

    Args:
        logit0: The log odds for the original proposition.
        logit1: The log odds for the negated proposition.

    Returns:
        The sum of the consistency and confidence losses.
    """
    p0, p1 = logit0.sigmoid(), logit1.sigmoid()

    consistency = p0.sub(1 - p1).square().mean()
    confidence = torch.min(p0, p1).square().mean()

    return consistency + confidence



"""Functions for extracting the hidden states of a model."""

from .prompt_dataset import Prompt, PromptDataset, PromptConfig
from ..utils import (
    assert_type,
    infer_label_column,
    select_usable_devices,
    float32_to_int16,
)
from .generator import _GeneratorBuilder
from dataclasses import dataclass, InitVar
from datasets import (
    Array3D,
    DatasetDict,
    Features,
    get_dataset_config_info,
    Sequence,
    Split,
    SplitDict,
    SplitInfo,
    Value,
)
from simple_parsing.helpers import field, Serializable
from transformers import (
    AutoConfig,
    AutoModel,
    AutoTokenizer,
    BatchEncoding,
    PreTrainedModel,
)
from typing import Iterable, Literal, Union
import logging
import torch


@dataclass
class ExtractionConfig(Serializable):
    """
    Args:
        model: HuggingFace model string identifying the language model to extract
            hidden states from.
        prompts: The configuration for the prompt prompts.
        layers: The layers to extract hidden states from.
        layer_stride: Shortcut for setting `layers` to `range(0, num_layers, stride)`.
        token_loc: The location of the token to extract hidden states from. Can be
            either "first", "last", or "mean". Defaults to "last".
        use_encoder_states: Whether to extract hiddens from the encoder in
            encoder-decoder models. Defaults to False.
    """

    prompts: PromptConfig
    model: str = field(positional=True)

    layers: tuple[int, ...] = ()
    layer_stride: InitVar[int] = 1
    token_loc: Literal["first", "last", "mean"] = "last"
    use_encoder_states: bool = False

    def __post_init__(self, layer_stride: int):
        if self.layers and layer_stride > 1:
            raise ValueError(
                "Cannot use both --layers and --layer-stride. Please use only one."
            )
        elif layer_stride > 1:
            from transformers import AutoConfig, PretrainedConfig

            # Look up the model config to get the number of layers
            config = assert_type(
                PretrainedConfig, AutoConfig.from_pretrained(self.model)
            )
            self.layers = tuple(range(0, config.num_hidden_layers, layer_stride))


def extract_hiddens(
    cfg: ExtractionConfig,
    *,
    device: Union[str, torch.device] = "cpu",
    rank: int = 0,
    split: str,
    world_size: int = 1,
) -> Iterable[dict]:
    """Run inference on a model with a set of prompts, yielding the hidden states.

    This is a lightweight, functional version of the `Extractor` API.
    """

    # Silence datasets logging messages from all but the first process
    if rank != 0:
        logging.disable(logging.CRITICAL)

    prompt_ds = PromptDataset(cfg.prompts, rank, world_size, split)
    if rank == 0:
        prompt_names = prompt_ds.prompter.all_template_names
        if cfg.prompts.num_variants >= 1:
            print(
                f"Using {cfg.prompts.num_variants} prompts per example: {prompt_names}"
            )
        elif cfg.prompts.num_variants == -1:
            print(f"Using all prompts per example: {prompt_names}")
        else:
            raise ValueError(f"Invalid prompt num_variants: {cfg.prompts.num_variants}")

    # AutoModel should do the right thing here in nearly all cases. We don't actually
    # care what head the model has, since we are just extracting hidden states.
    model = AutoModel.from_pretrained(cfg.model, torch_dtype="auto").to(device)
    # TODO: Maybe also make this configurable?
    # We want to make sure the answer is never truncated
    tokenizer = AutoTokenizer.from_pretrained(cfg.model, truncation_side="left")

    if cfg.use_encoder_states and not model.config.is_encoder_decoder:
        raise ValueError(
            "use_encoder_states is only compatible with encoder-decoder models."
        )

    # TODO: Make this configurable or something
    # Token used to separate the question from the answer
    num_choices = prompt_ds.num_classes
    sep_token = tokenizer.sep_token or "\n"

    if not tokenizer.pad_token:
        tokenizer.pad_token = tokenizer.eos_token

    # Whether to concatenate the question and answer before passing to the model.
    # If False pass them to the encoder and decoder separately.
    is_enc_dec = model.config.is_encoder_decoder
    should_concat = not is_enc_dec or cfg.use_encoder_states

    def tokenize(prompt: Prompt, idx: int, **kwargs):
        return tokenizer(
            (
                [prompt.to_string(idx, sep=sep_token)]
                if should_concat
                else [prompt.question]
            ),
            text_target=(
                [prompt.answers[idx]] if not should_concat else None
            ),  # type: ignore
            padding=True,
            return_tensors="pt",
            truncation=True,
            **kwargs,
        ).to(device)

    # This function returns the flattened questions and answers. After inference we
    # need to reshape the results.
    def collate(prompts: list[Prompt]) -> list[list[BatchEncoding]]:
        return [[tokenize(prompt, i) for i in range(num_choices)] for prompt in prompts]

    # If this is an encoder-decoder model and we're passing the answer to the encoder,
    # we don't need to run the decoder at all. Just strip it off, making the problem
    # equivalent to a regular encoder-only model.
    if is_enc_dec and cfg.use_encoder_states:
        # This isn't actually *guaranteed* by HF, but it's true for all existing models
        if not hasattr(model, "get_encoder") or not callable(model.get_encoder):
            raise ValueError(
                "Encoder-decoder model doesn't have expected get_encoder() method"
            )

        model = assert_type(PreTrainedModel, model.get_encoder())

    # Iterating over questions
    layer_indices = cfg.layers or tuple(range(model.config.num_hidden_layers))
    for prompts in prompt_ds:
        inputs = collate(prompts)
        hidden_dict = {
            f"hidden_{layer_idx}": torch.empty(
                prompt_ds.num_variants,
                num_choices,
                model.config.hidden_size,
                device=device,
                dtype=torch.int16,
            )
            for layer_idx in layer_indices
        }
        variant_ids = [prompt.template_name for prompt in prompts]

        # Iterate over variants
        for i, variant_inputs in enumerate(inputs):
            # Iterate over answers
            for j, inpt in enumerate(variant_inputs):
                outputs = model(**inpt, output_hidden_states=True)

                hiddens = (
                    outputs.get("decoder_hidden_states") or outputs["hidden_states"]
                )
                # First element of list is the input embeddings
                hiddens = hiddens[1:]

                # Throw out layers we don't care about
                hiddens = [hiddens[i] for i in layer_indices]

                # Current shape of each element: (batch_size, seq_len, hidden_size)
                if cfg.token_loc == "first":
                    hiddens = [h[..., 0, :] for h in hiddens]
                elif cfg.token_loc == "last":
                    hiddens = [h[..., -1, :] for h in hiddens]
                elif cfg.token_loc == "mean":
                    hiddens = [h.mean(dim=-2) for h in hiddens]
                else:
                    raise ValueError(f"Invalid token_loc: {cfg.token_loc}")

                for layer_idx, hidden in zip(layer_indices, hiddens):
                    hidden_dict[f"hidden_{layer_idx}"][i, j] = float32_to_int16(hidden)

        yield dict(
            label=prompts[0].label,
            variant_ids=variant_ids,
            **hidden_dict,
        )


# Dataset.from_generator wraps all the arguments in lists, so we unpack them here
def _extraction_worker(**kwargs):
    yield from extract_hiddens(**{k: v[0] for k, v in kwargs.items()})


def extract(cfg: ExtractionConfig, max_gpus: int = -1) -> DatasetDict:
    """Extract hidden states from a model and return a `DatasetDict` containing them."""

    def get_splits() -> SplitDict:
        base_splits = assert_type(SplitDict, info.splits)
        splits = set(base_splits) & {Split.TRAIN, Split.VALIDATION, Split.TEST}
        if Split.VALIDATION in splits and Split.TEST in splits:
            splits.remove(Split.TEST)

        limit = cfg.prompts.max_examples or int(1e100)
        return SplitDict(
            {
                k: SplitInfo(
                    name=k,
                    num_examples=min(limit, v.num_examples),
                    dataset_name=v.dataset_name,
                )
                for k, v in base_splits.items()
                if k in splits
            },
            dataset_name=base_splits.dataset_name,
        )

    model_cfg = AutoConfig.from_pretrained(cfg.model)
    num_variants = cfg.prompts.num_variants
    ds_name, _, config_name = cfg.prompts.dataset.partition(" ")
    info = get_dataset_config_info(ds_name, config_name or None)

    features = assert_type(Features, info.features)
    label_col = cfg.prompts.label_column or infer_label_column(features)

    splits = get_splits()

    layer_cols = {
        f"hidden_{layer}": Array3D(
            dtype="int16",
            shape=(num_variants, 2, model_cfg.hidden_size),
        )
        for layer in cfg.layers or range(model_cfg.num_hidden_layers)
    }
    other_cols = {
        "variant_ids": Sequence(
            Value(dtype="string"),
            length=num_variants,
        ),
        "label": features[label_col],
    }
    devices = select_usable_devices(max_gpus)
    builders = {
        split_name: _GeneratorBuilder(
            cache_dir=None,
            features=Features({**layer_cols, **other_cols}),
            generator=_extraction_worker,
            split_name=split_name,
            split_info=split_info,
            gen_kwargs=dict(
                cfg=[cfg] * len(devices),
                device=devices,
                rank=list(range(len(devices))),
                split=[split_name] * len(devices),
                world_size=[len(devices)] * len(devices),
            ),
        )
        for (split_name, split_info) in splits.items()
    }

    ds = dict()
    for split, builder in builders.items():
        # TODO: we may need a barrier here?
        builder.download_and_prepare(num_proc=len(devices))
        ds[split] = builder.as_dataset(split=split)
    return DatasetDict(ds)



from .data_utils import (
    compute_class_balance,
    get_columns_all_equal,
    held_out_split,
    infer_label_column,
    undersample,
    float32_to_int16,
    int16_to_float32,
)
from .gpu_utils import select_usable_devices
from .tree_utils import pytree_map
from .typing import assert_type



from typing import cast, Any, Type, TypeVar


T = TypeVar("T")


def assert_type(typ: Type[T], obj: Any) -> T:
    """Assert that an object is of a given type at runtime and return it."""
    if not isinstance(obj, typ):
        raise TypeError(f"Expected {typ.__name__}, got {type(obj).__name__}")

    return cast(typ, obj)



from dataclasses import dataclass, field
from typing import Callable, Optional, Any, Dict

import datasets
from datasets.splits import NamedSplit


@dataclass
class _GeneratorConfig(datasets.BuilderConfig):
    generator: Optional[Callable] = None
    gen_kwargs: dict[str, Any] = field(default_factory=dict)
    features: Optional[datasets.Features] = None


@dataclass
class _SplitGenerator:
    """
    datasets.SplitGenerator but use given `split_info` instead initializing a new one
    """

    name: str
    split_info: datasets.SplitInfo
    gen_kwargs: Dict = field(default_factory=dict)

    def __post_init__(self):
        self.name = str(self.name)  # Make sure we convert NamedSplits in strings
        NamedSplit(self.name)  # check that it's a valid split name


class _GeneratorBuilder(datasets.GeneratorBasedBuilder):
    """Patched version of `datasets.Generator` allowing for splits besides `train`"""

    BUILDER_CONFIG_CLASS = _GeneratorConfig
    config: _GeneratorConfig

    def __init__(self, split_name: str, split_info: datasets.SplitInfo, **kwargs):
        self.split_name = split_name
        self.split_info = split_info

        super().__init__(**kwargs)

    def _info(self):
        return datasets.DatasetInfo(features=self.config.features)

    def _split_generators(self, dl_manager):
        return [
            _SplitGenerator(
                name=self.split_name,
                split_info=self.split_info,
                gen_kwargs=self.config.gen_kwargs,
            )
        ]

    def _generate_examples(self, **gen_kwargs):
        assert self.config.generator is not None, "generator must be specified"
        for idx, ex in enumerate(self.config.generator(**gen_kwargs)):
            yield idx, ex



"""Main entry point for `elk`."""

from .extraction import extract, ExtractionConfig
from .training import RunConfig
from .training.train import train
from pathlib import Path
from simple_parsing import ArgumentParser


def run():
    parser = ArgumentParser(add_help=False)
    subparsers = parser.add_subparsers(dest="command", required=True)

    extract_parser = subparsers.add_parser(
        "extract",
        help="Extract hidden states from a model.",
    )
    extract_parser.add_arguments(ExtractionConfig, dest="extraction")
    extract_parser.add_argument(
        "--output",
        "-o",
        type=Path,
        help="Path to save hidden states to.",
        required=True,
    )
    extract_parser.add_argument(
        "--max_gpus",
        type=int,
        help="Maximum number of GPUs to use.",
        required=False,
        default=-1,
    )

    elicit_parser = subparsers.add_parser(
        "elicit",
        help=(
            "Extract and train a set of ELK reporters "
            "on hidden states from `elk extract`. "
        ),
        conflict_handler="resolve",
    )
    elicit_parser.add_arguments(RunConfig, dest="run")
    elicit_parser.add_argument(
        "--output",
        "-o",
        type=Path,
        help="Path to save checkpoints to.",
    )

    subparsers.add_parser(
        "eval", help="Evaluate a set of ELK reporters generated by `elk train`."
    )
    args = parser.parse_args()

    if args.command == "extract":
        extract(args.extraction, args.max_gpus).save_to_disk(args.output)
    elif args.command == "elicit":
        train(args.run, args.output)

    elif args.command == "eval":
        # TODO: Implement evaluation script
        raise NotImplementedError
    else:
        raise ValueError(f"Unknown command {args.command}")


if __name__ == "__main__":
    run()



"""Utilities that use PyNVML to get GPU usage info, and select GPUs accordingly."""

import os
import pynvml
import torch
import warnings


def select_usable_devices(max_gpus: int = -1, *, min_memory: int = 0) -> list[str]:
    """Select a set of devices that have at least `min_memory` bytes of free memory.

    When there are more than enough GPUs to satisfy the request, the GPUs with the
    most free memory will be selected. With default arguments, this function will
    simply return a list of all available GPU indices. A human-readable message will
    be printed to stdout indicating which GPUs were selected.

    This function uses the PyNVML library to get usage info. Unfortunately, PyNVML does
    not respect the `CUDA_VISIBLE_DEVICES` environment variable, while PyTorch does.
    Correctly converting PyNVML device indices to PyTorch indices is nontrivial and was
    only recently (commit `dc4f2af` on 9 Feb. 2023) implemented in PyTorch `master`. We
    can't depend on PyTorch nightly and we also don't want to copy-paste the code here.

    For now, we simply return `list(range(max_gpus))` whenever `CUDA_VISIBLE_DEVICES`
    is set. Arguably this is expected behavior. If the user set `CUDA_VISIBLE_DEVICES`,
    they probably want to use all & only those GPUs.

    Args:
        num_gpus: Maximum number of GPUs to select. If negative, all available GPUs
            meeting the criteria will be selected.
        min_memory: Minimum amount of free memory (in bytes) required to select a GPU.

    Returns:
        A list of suitable PyTorch device strings, in ascending numerical order.

    Raises:
        ValueError: If `max_gpus` is greater than the number of visible GPUs.
    """
    # Trivial case: no GPUs requested or available
    num_visible = torch.cuda.device_count()
    if max_gpus == 0 or num_visible == 0:
        return ["cpu"]

    # Sanity checks
    if max_gpus > num_visible:
        raise ValueError(
            f"Requested {max_gpus} GPUs, but only {num_visible} are visible."
        )
    elif max_gpus < 0:
        max_gpus = num_visible

    # No limits, so try to use all installed GPUs
    if max_gpus == num_visible and min_memory <= 0:
        print(f"Using all {num_visible} GPUs.")
        return [f"cuda:{i}" for i in range(max_gpus)]

    # The user set CUDA_VISIBLE_DEVICES and also requested a specific number of GPUs.
    # The environment variable takes precedence, so we'll just use all visible GPUs.
    count_msg = "all" if max_gpus == num_visible else f"first {max_gpus}"
    if "CUDA_VISIBLE_DEVICES" in os.environ:
        warnings.warn(
            f"Smart GPU selection not supported when CUDA_VISIBLE_DEVICES is set. "
            f"Will use {count_msg} visible devices."
        )
        return [f"cuda:{i}" for i in range(max_gpus)]

    # pynvml.nvmlInit() will raise if we're using non-NVIDIA GPUs
    try:
        pynvml.nvmlInit()
    except pynvml.NVMLError:
        warnings.warn(
            f"Unable to initialize PyNVML; are you using non-NVIDIA GPUs? Will use "
            f"{count_msg} visible devices."
        )
        return [f"cuda:{i}" for i in range(max_gpus)]

    try:
        # PyNVML and PyTorch device indices should agree when CUDA_VISIBLE_DEVICES is
        # not set. We need them to agree so that the PyNVML indices match the PyTorch
        # indices, and we don't have to do any complex error-prone conversions.
        num_installed = pynvml.nvmlDeviceGetCount()
        assert num_installed == num_visible, "PyNVML and PyTorch disagree on GPU count"

        # List of (-free memory, GPU index) tuples. Sorted descending by free memory,
        # then ascending by GPU index.
        memories_and_indices = sorted(
            (
                -int(pynvml.nvmlDeviceGetMemoryInfo(handle).free),
                pynvml.nvmlDeviceGetIndex(handle),
            )
            for handle in map(pynvml.nvmlDeviceGetHandleByIndex, range(num_installed))
        )
        usable_indices = [
            index for neg_mem, index in memories_and_indices if -neg_mem >= min_memory
        ]
    finally:
        # Make sure we always shut down PyNVML
        pynvml.nvmlShutdown()

    # Indices are sorted descending by free memory, so we want the first `max_gpus`
    # items. For printing purposes, though, we sort the indices numerically.
    selection = sorted(usable_indices[:max_gpus])

    # Did we get the maximum number of GPUs requested?
    if len(selection) == max_gpus:
        print(f"Using {len(selection)} of {num_visible} GPUs: {selection}")
    else:
        print(f"Using {len(selection)} of {max_gpus} requested GPUs: {selection}")
        print(
            f"{num_visible - len(selection)} GPUs have insufficient free memory "
            f"({min_memory / 10 ** 9:.2f} GB needed)."
        )

    if len(selection) > 0:
        return [f"cuda:{i}" for i in selection]
    else:
        return ["cpu"]



from ..math import stochastic_round_constrained
from ..utils import assert_type, compute_class_balance, infer_label_column, undersample
from dataclasses import dataclass
from datasets import DatasetDict, load_dataset
from numpy.typing import NDArray
from promptsource_module.promptsource.templates import DatasetTemplates
from random import Random
from simple_parsing.helpers import field, Serializable
from torch.utils.data import Dataset as TorchDataset
from typing import Optional
import numpy as np


@dataclass
class Prompt:
    """A question, its possible answers, and the correct answer.
    A question is a template applied to a predicate."""

    question: str
    answers: list[str]
    label: int
    template_name: str

    def to_string(self, answer_idx: int, sep: str = "\n") -> str:
        return f"{self.question}{sep}{self.answers[answer_idx]}"


@dataclass
class PromptConfig(Serializable):
    """
    Args:
        dataset: Space-delimited name of the HuggingFace dataset to use, e.g.
            `"super_glue boolq"` or `"imdb"`.
        balance: Whether to force class balance in the dataset using undersampling.
        label_column: The column containing the labels. By default, we infer this from
            the datatypes of the columns in the dataset; if there is only one column
            with a `ClassLabel` datatype, we use that.
        max_examples: The maximum number of examples to use from the dataset. If zero,
            use all examples. Defaults to 0.
        num_shots: The number of examples to use in few-shot prompts. If zero, prompts
            are zero-shot. Defaults to 0.
        seed: The seed to use for prompt randomization. Defaults to 42.
        num_variants: The number of prompt templates to apply to each predicate upon
            call to __getitem__. Use -1 to apply all available templates. Defaults to 1.
    """

    dataset: str = field(positional=True)
    balance: bool = False
    label_column: Optional[str] = None
    max_examples: int = 0
    num_shots: int = 0
    seed: int = 42
    num_variants: int = 1


class PromptDataset(TorchDataset):
    """Wrapper for a HuggingFace dataset which generates prompts with `promptsource`.

    Usually `promptsource` has multiple prompt templates for a given dataset. We handle
    this in two ways. When `strategy` is set to `"randomize"` (the default), we sample
    a random prompt template for each example on-the-fly when `__getitem__` is called,
    using the seed passed to `__init__`. Note this means that the same example may be
    assigned a different prompt template if `__getitem__` is called multiple times with
    the same index.  TODO redo this documentation

    When `strategy` is set to `"all"`, we "broadcast" the prompt templates across the
    dataset, multiplying its effective size by the number of templates.

    Example:
    >>> prompts = PromptDataset("super_glue", "boolq", split="train")
    >>> prompt = prompts[0]
    >>> prompt.to_string(0)
    "Henry Mills (Once Upon a Time) -- Henry Daniel Mills is a fictional character...
    """

    def __init__(
        self,
        cfg: PromptConfig,
        rank: int = 0,
        world_size: int = 1,
        split: str = "validation",
    ):
        ds_name, _, config_name = cfg.dataset.partition(" ")

        self.num_shots = cfg.num_shots
        self.prompter = DatasetTemplates(ds_name, config_name or None)  # type: ignore
        self.rng = Random(cfg.seed)
        self.num_variants = (
            cfg.num_variants if cfg.num_variants > 0 else len(self.prompter.templates)
        )

        ds_dict = assert_type(
            DatasetDict,  # TODO: Should we support IterableDataset?
            load_dataset(ds_name, config_name or None),
        )

        # By default we use the existing train-validation/test split in the dataset.
        # If it doesn't exist, we create our own 75/25 train-test split. Crucially,
        # because the RNG is always seeded, this split will be the same for independent
        # instantiations of PromptDataset (unless you set the seed to something else).
        # This allows you to just set split="train" and split="test" for any dataset
        # and not worry about train-test leakage.
        split_name, *others = ds_dict.keys()
        if not others:
            print("Creating a 75/25 train-test split...")

            # Don't shuffle now because we're going to shuffle later
            ds_dict = ds_dict[split_name].train_test_split(
                seed=cfg.seed, shuffle=False, stratify_by_column=cfg.label_column
            )
            assert isinstance(ds_dict, DatasetDict)

        # The 'active' split is the one that gets queried by __getitem__
        self.active_split = ds_dict[split]
        label_col = cfg.label_column or infer_label_column(self.active_split.features)
        self.label_column = label_col

        # Enforce class balance if needed
        if cfg.balance:
            self.active_split = undersample(self.active_split, self.rng, label_col)
            self.class_fracs = np.ones(self.num_classes) / self.num_classes
        else:
            class_sizes = compute_class_balance(self.active_split, label_col)
            self.class_fracs: NDArray[np.floating] = class_sizes / class_sizes.sum()

        # We use stratified sampling to create few-shot prompts that are as balanced as
        # possible. If needed, create the strata now so that we can use them later.
        if cfg.num_shots > 0:
            # Sanity check that we can fit an example from every class in the prompt
            if self.num_classes > cfg.num_shots:
                raise ValueError(
                    f"Few-shot prompts should contain at least one example from each "
                    f"class; got {cfg.num_shots} examples, {self.num_classes} classes"
                )

            # Sanity check to prevent train-test leakage via few-shot prompts
            if "train" not in ds_dict:
                raise ValueError(
                    f"Dataset {cfg.dataset} has no train split, so we can't create "
                    "few-shot prompts"
                )

            self.fewshot_strata = [
                ds_dict["train"].filter(lambda ex: ex[label_col] == i)
                for i in range(self.num_classes)
            ]
        else:
            self.fewshot_strata = []

        # Now shuffle the active split and truncate it if needed
        self.active_split = self.active_split.shuffle(seed=cfg.seed)
        if 0 < cfg.max_examples < len(self.active_split):
            self.active_split = self.active_split.select(range(cfg.max_examples))

        # Shard if needed
        if world_size > 1:
            self.active_split = self.active_split.shard(world_size, rank)

    def __getitem__(self, index: int) -> list[Prompt]:
        """Get a list of prompts for a given predicate"""
        # get self.num_variants unique prompts from the template pool
        template_names = self.rng.sample(
            self.prompter.templates.keys(), self.num_variants
        )

        example = self.active_split[index]

        prompts = []
        for template_name in template_names:
            template = self.prompter.templates[template_name]

            true_label = example[self.label_column]
            answers = []
            questions = set()

            for fake_label in range(self.num_classes):
                example[self.label_column] = fake_label

                q, a = template.apply(example)
                answers.append(a)
                questions.add(q)

            assert len(questions) == 1
            question = questions.pop()

            if self.num_shots > 0:
                # Use stratified sampling to get `num_shots` examples from train set.
                # If `num_shots` is not divisible by the number of classes, stochastic
                # rounding is used to determine the number of examples per class.
                example_counts = stochastic_round_constrained(
                    list(self.class_fracs * self.num_shots), self.rng
                )
                examples = []

                for count, stratum in zip(example_counts, self.fewshot_strata):
                    indices = self.rng.sample(range(len(stratum)), count)

                    for idx in indices:
                        q, a = template.apply(stratum[idx])
                        examples.append(f"{q}\n{a}")

                self.rng.shuffle(examples)
                question = "\n\n".join(examples + [question])

            prompts.append(
                Prompt(
                    question=question,
                    answers=answers,
                    label=true_label,
                    template_name=template_name,
                )
            )
        return prompts

    def __iter__(self):
        return (self[i] for i in range(len(self.active_split)))

    def __len__(self):
        """Get the number of predicates in the dataset."""
        return len(self.active_split)

    @property
    def num_classes(self) -> int:
        """Number of classes in the underlying dataset."""

        # We piggyback on the ClassLabel feature type to get the number of classes
        return self.active_split.features[self.label_column].num_classes



"""Preprocessing functions for training."""

from typing import Literal
import torch


def normalize(
    train_hiddens: torch.Tensor,
    val_hiddens: torch.Tensor,
    method: Literal["legacy", "none", "elementwise", "meanonly"] = "legacy",
) -> tuple[torch.Tensor, torch.Tensor]:
    """Normalize the hidden states.

    Normalize the hidden states with the specified method.
    The "legacy" method is the same as the original ELK implementation.
    The "elementwise" method normalizes each element.
    The "meanonly" method normalizes the mean.

    Args:
        train_hiddens: The hidden states for the training set.
        val_hiddens: The hidden states for the validation set.
        method: The normalization method to use.

    Returns:
        tuple containing the training and validation hidden states.
    """
    if method == "none":
        return train_hiddens, val_hiddens
    elif method == "legacy":
        master = torch.cat([train_hiddens, val_hiddens], dim=0).float()
        means = master.mean(dim=0)

        train_hiddens -= means
        val_hiddens -= means

        scale = master.shape[-1] ** 0.5 / master.norm(dim=-1).mean()
        train_hiddens *= scale
        val_hiddens *= scale
    else:
        means = train_hiddens.float().mean(dim=0)
        train_hiddens -= means
        val_hiddens -= means

        if method == "elementwise":
            scale = 1 / train_hiddens.norm(dim=0, keepdim=True)
        elif method == "meanonly":
            scale = 1
        else:
            raise NotImplementedError(f"Scale method '{method}' is not supported.")

        train_hiddens *= scale
        val_hiddens *= scale

    return train_hiddens, val_hiddens
